# Copyright (c) 2015-2016, The Authors and Contributors
# <see AUTHORS file>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
# following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
# following disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


from typing import Tuple

from supplychainpy.bi._recommendation_state_machine import RecommendationStateMachine


class DashMachine(RecommendationStateMachine):
    def __init__(self):
        super(DashMachine, self).__init__()

    def run(self, message: str) -> Tuple[str]:
        new_state = ''
        try:
            handler = self.handlers[self.start_state]
        except:
            raise OSError("call .set_start() first before .run()")
        if not self.end_states:
            raise OSError("must have at least one end_state")

        while True:
            (new_state, message) = handler(message)
            if new_state.upper() in self.end_states:
                break
            else:
                handler = self.handlers[new_state.upper()]
        return new_state, message



# Copyright (c) 2015-2016, The Authors and Contributors
# <see AUTHORS file>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
# following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
# following disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from decimal import Decimal
from warnings import warn
import csv
import logging

from supplychainpy._helpers._decorators import log_this

log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())

TABLE_HEADINGS = {
    'UNIT_COST': 'unit_cost',
    'LEAD_TIME': 'lead_time',
    'RETAIL_PRICE': 'retail_price',
    'QUANTITY_ON_HAND': 'quantity_on_hand',
    'BACKLOG': 'backlog'
}

# TODO-feature format pandas frame
# TODO-feature allow data munger to accept any delimiter as a parameter
def clean_orders_data_col_txt(file) -> dict:
    """

    Args:
        file:

    Returns:
        dict:
    """
    item_list = {}
    split_line = []
    for line in file:
        split_line.append(line.split("|"))
    for item in split_line:
        item_list[item[0]] = Decimal(item[1].strip())
    return item_list


# TODO-feature make csv version of clean_orders also take into account column number add tests (create csv file)
def clean_orders_data_col_csv(file) -> dict:
    item_list = {}
    read_csv = csv.reader(file)
    headers = next(read_csv)
    split_line = list(read_csv)
    for item in split_line:
        item_list[item[0]] = Decimal(item[1])
    return item_list


# TODO-feature allow data munger to accept any delimiter as a parameter
def clean_orders_data_row(file, length: int) -> list:
    collection = []
    try:
        sku_list = []
        split_line = []
        composite = {}

        # unpacks the line form the txt file and splits using delimiter (default pipe) to split_line list
        for line in file:
            split_line.append(line.split("|"))
        # sorts all parts of the list into labeled dictionary
        for row in split_line:
            for index, item in enumerate(row):
                if int(index) < 1:
                    composite["sku_id"] = item
                elif 1 <= int(index) <= length:
                    sku_list.append(item.strip("\n"))
                elif int(index) == length + 1:
                    composite["unit_cost"] = item.strip("\n")
                elif int(index) == length + 2:
                    composite["lead_time"] = item.strip("\n")
                    # if the sku id is not unique the the data will just append. Need to check if there are duplicates
                    # and throw an exception.
                    composite["demand"] = sku_list
                    sku_list = []
                    collection.append(composite)
                    composite = {}
    except OSError:
        print("")

    return collection


# TODO-feature cleans a text or csv file and insert into numpy array
# remember to specify in documentation that the orders data will assume 12 months unless otherwise stated

@log_this(logging.INFO, message='clean orders data.')
def clean_orders_data_row_csv(file, length: int = 12) -> list:
    # check length of list, check length of the length specified and validate the length and make sure it is long enough
    # to support the unit_cost, lead_time, asp, quantity_on_hand
    collection = []
    try:

        sku_list = []
        composite = {}

        read_csv = csv.reader(file)
        headers = next(read_csv)
        split_line = list(read_csv)

        for row in split_line:
            if length == len(row) - 6:
                for index, item in enumerate(row):
                    if int(index) < 1:
                        composite["sku_id"] = item
                    elif 1 <= index <= length:
                        sku_list.append(item)
                    elif index == length + 1:
                        composite["unit_cost"] = item
                        log.info("Extracted unit cost for sku. SKU: {} UNIT COST: {} ".format(
                            composite.get("sku_id", "UNKNOWN_SKU"), item))
                    elif index == length + 2:
                        composite["lead_time"] = item
                        log.info("Extracted unit cost for sku. SKU: {} LEAD-TIME: {} ".format(
                            composite.get("sku_id", "UNKNOWN_SKU"), item))
                    elif index == length + 3:
                        composite["retail_price"] = item
                        log.info("Extracted unit cost for sku. SKU: {}  RETAIL PRICE: {} ".format(
                            composite.get("sku_id", "UNKNOWN_SKU"), item))
                    elif index == length + 4:
                        composite["quantity_on_hand"] = item
                        log.info("Extracted unit cost for sku. SKU: {} QUANTITY ON HAND: {} ".format(
                            composite.get("sku_id", "UNKNOWN_SKU"), item))
                    elif index == length + 5:
                        composite["backlog"] = item
                        log.info("Extracted unit cost for sku. SKU: {} BACKLOG: {} ".format(
                            composite.get("sku_id", "UNKNOWN_SKU"), item))
                        # if the sku id is not unique the the data will just append. Need to check if there are duplicates
                        # and through an exception.
                composite["demand"] = tuple(sku_list)
                composite["headers"] = headers
                sku_list = []

                if composite.get("sku_id") is None or composite.get("unit_cost") is None or composite.get("lead_time") is \
                        None or composite.get("retail_price") is None:
                    err_msg = "csv file is formatted incorrectly. Please make sure the  formatted file\n [sku_id," \
                              " orders1,orders2....unit_cost, lead_time, retail_price, quantity_on_hand"
                    raise Exception(err_msg)

                collection.append(composite)
                composite = {}
            else:
                formatting_err = "The file formatting is incorrect. The specified column count supplied as a" \
                                 " parameter is {}.\n Including the sku_id, unit_cost, lead_time, retail_price\n " \
                                 "and quantity on hand the csv row should be {} columns long.\n The current" \
                                 " column count is {}. Please check the file or specified length."
                raise Exception(formatting_err.format(length, (length + 1) + len(TABLE_HEADINGS), len(row)))
    except OSError as e:
        print(e)

    return collection

def match_headers(headers: list)->dict:
    pass

def check_extension(file_path, file_type: str) -> bool:
    """ Check the correct file type has been selected.

    Args:
        file_path (file):   The path to the file containing two columns of data, 1 period and 1 data-point for 1 sku.
        file_type (str):    specifying 'csv' or 'text'
    Returns:
        bool:   Extension present or not.

    """
    try:
        if file_path.endswith(".txt") and file_type.lower() == "text":
            flag = True
        elif file_path.endswith(".csv") and file_type.lower() == "csv":
            flag = True
        else:
            print(file_path, file_type)
            flag = False
        return flag
    except OSError as e:
        print(file_path, file_type ,e)


# refocator length to column count
# if a user specifies a lower column_count than actually supplied then the oher columns are going to be incorrect.
# probably best to check heading using regx



# Copyright (c) 2015-2016, The Authors and Contributors
# <see AUTHORS file>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
# following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
# following disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import itertools
import logging
from collections import Iterable
from decimal import Decimal
from decimal import getcontext

from supplychainpy._helpers._decorators import log_this
from supplychainpy._helpers._enum_formats import PeriodFormats
from supplychainpy.model_demand import holts_trend_corrected_exponential_smoothing_forecast
from supplychainpy.model_demand import simple_exponential_smoothing_forecast


def _standard_deviation_orders(orders: dict, average_order: Decimal) -> Decimal:
    deviation = Decimal(0)
    variance = []
    for x in orders.values():
        variance.append(Decimal(x - average_order))
    for j in variance:
        deviation += Decimal(j) ** Decimal(2)
    deviation /= Decimal(len(variance))
    return Decimal(Decimal(deviation) ** Decimal(0.5))


# TODO-feature convert orders data into correct period
# TODO-feature convert list of lead-times to average lead-times

class UncertainDemand:
    """ Models inventory profile calculating economic order quantity, variable cost, reorder quantity and
        ABCXYZ classification
    """
    __z_value = Decimal(0.00)  # default set to 90%
    __lead_time = 0
    __safety_stock = 0
    __demand_variability = Decimal(0)
    __reorder_level = 0
    __unit_cost = Decimal(0.00)
    __reorder_cost = Decimal(00.00)
    __CONST_HOLDING_COST_FACTOR = Decimal(0.25)
    __fixed_reorder_quantity = 0
    __DAYS = Decimal(7)
    __WEEKS = Decimal(4)
    __MONTHS = Decimal(12)
    __QUARTER = Decimal(4)
    __period = 0
    __sku_revenue = Decimal(0)
    __percentage_of_revenue = Decimal(0)
    __cumulative_percentage = Decimal(0)
    __abc_classification = ""
    __xyz_classification = ""
    __economic_order_variable_cost = Decimal(0)
    __economic_order_qty = Decimal(0)
    getcontext().prec = 8
    _summary_keywords = ['sku', 'standard_deviation', 'safety_stock', 'demand_variability', 'reorder_level',
                         'reorder_quantity', 'revenue', 'economic_order_quantity', 'economic_order_variable_cost',
                         'ABC_XYZ_Classification', 'excess_stock', 'shortages', 'average_orders', 'unit_cost',
                         'quantity_on_hand', 'currency', 'orders', 'total_orders', 'backlog']
    __rank = 0

    def __init__(self, orders: dict, sku: str, currency: str, lead_time: Decimal, unit_cost: Decimal,
                 reorder_cost: Decimal,
                 z_value: Decimal = Decimal(1.28), holding_cost: Decimal = 0.00, retail_price: Decimal = 0.00,
                 period: str = PeriodFormats.months.name, quantity_on_hand: Decimal = 0.00, backlog: Decimal = 0.00):
        self.__currency = currency
        self.__orders = orders
        self.__sku_id = sku
        self.__lead_time = Decimal(lead_time)
        self.__unit_cost = Decimal(unit_cost)
        self.__retail_price = Decimal(retail_price)
        self.__quantity_on_hand = Decimal(quantity_on_hand)
        self.__z_value = z_value
        self.__count_orders = len(self.__orders)

        if len(orders) < 2:
            self.__average_order = Decimal(self.average_order_row())
            self.__orders_standard_deviation = self._standard_deviation_orders_row()
        else:
            self.__average_order = Decimal(self.average_order())
            self.__orders_standard_deviation = _standard_deviation_orders(orders=orders,
                                                                          average_order=self.__average_order)
        self.__sku_revenue = self._revenue(orders=orders)
        self.__safety_stock = self._safety_stock()
        self.__demand_variability = self._demand_variability()
        self.__reorder_level = Decimal(self._reorder_level())
        self.__reorder_cost = Decimal(reorder_cost)
        self.__fixed_reorder_quantity = Decimal(self._fixed_order_quantity())
        self.__period = period
        self.__excess_stock = self._excess_qty()
        self.__backlog = Decimal(backlog) if float(backlog) > 0 else 0
        self.__shortage_qty = self._shortage_qty()
        self.__total_orders = self._sum_orders()


    @property
    def backlog(self) -> Decimal:
        return self.__backlog

    @property
    def simple_exponential_smoothing_forecast(self):
        return self._generate_optimised_ses_forecast()

    @property
    def holts_trend_corrected_forecast(self):
        return self._generate_holts_es_forecast()

    @property
    def total_orders(self):
        return self.__total_orders

    @property
    def currency(self):
        return self.__currency

    @property
    def quantity_on_hand(self):
        return self.__quantity_on_hand

    @property
    def excess_stock_cost(self):
        return self.__excess_stock * self.__unit_cost

    @property
    def shortage_cost(self):
        return self.__shortage_qty * self.__unit_cost

    @property
    def safety_stock_cost(self):
        return self.__safety_stock * self.__unit_cost

    @property
    def retail_price(self):
        return self.__retail_price

    @retail_price.setter
    def retail_price(self, retail_price):
        self.__retail_price = retail_price

    @property
    def excess_stock(self):
        return self.__excess_stock

    @excess_stock.setter
    def excess_stock(self, excess):
        self.__excess_stock = excess

    @property
    def shortages(self):
        return self.__shortage_qty

    @shortages.setter
    def shortages(self, shortage):
        self.__shortage_qty = shortage

    @property
    def safety_stock(self):
        return self.__safety_stock

    @safety_stock.setter
    def safety_stock(self, safety_stock):
        self.__safety_stock = safety_stock

    @property
    def reorder_level(self):
        return self.__reorder_level

    @reorder_level.setter
    def reorder_level(self, reorder_level):
        self.__reorder_level = reorder_level

    @property
    def unit_cost(self) -> Decimal:
        return self.__unit_cost

    @unit_cost.setter
    def unit_cost(self, unit_cost):
        self.__unit_cost = unit_cost

    @property
    def abcxyz_classification(self) -> str:
        """Gets ABCXYZ classification as a concatenated string"""
        return self.__abc_classification + self.__xyz_classification

    @property
    def abc_classification(self) -> str:
        return self.__abc_classification

    @abc_classification.setter
    def abc_classification(self, abc_classifier: str):
        self.__abc_classification = abc_classifier

    @property
    def xyz_classification(self) -> str:
        return self.__xyz_classification

    @xyz_classification.setter
    def xyz_classification(self, xyz_classifier: str):
        self.__xyz_classification = xyz_classifier

    @property
    def percentage_revenue(self) -> Decimal:
        return self.__percentage_of_revenue

    @percentage_revenue.setter
    def percentage_revenue(self, percentage_orders):
        self.__percentage_of_revenue = percentage_orders

    @property
    def cumulative_percentage(self) -> Decimal:
        return self.__cumulative_percentage

    @cumulative_percentage.setter
    def cumulative_percentage(self, percentage_orders):
        self.__cumulative_percentage = percentage_orders

    @property
    def orders(self):
        return self.__orders.get("demand")

    @property
    def order(self):
        total_order = 0
        orders_list = []
        for items in self.__orders:
            orders_list = self.__orders[items]
        if isinstance(orders_list, Iterable):
            for item in orders_list:
                total_order += Decimal(item)
        else:
            for item in self.__orders:
                total_order += self.__orders[item]
        return total_order

    @property
    def order_count(self):
        return self.__count_orders

    @order.setter
    def order(self, orders):
        self.__orders = orders

    @property
    def sku_id(self):
        return self.__sku_id

    @sku_id.setter
    def sku_id(self, sku):
        self.__sku_id = sku

    @property
    def lead_time(self):
        return self.__lead_time

    @lead_time.setter
    def lead_time(self, lead_time):
        self.__lead_time = lead_time

    @property
    def average_orders(self):
        return self.__average_order

    @property
    def standard_deviation(self):
        return self.__orders_standard_deviation

    @property
    def revenue(self) -> Decimal:
        return self.__sku_revenue

    @property
    def demand_variability(self) -> Decimal:
        return self._demand_variability()

    @demand_variability.setter
    def demand_variability(self, demand_variability: Decimal):
        self.__demand_variability = demand_variability

    @property
    def fixed_order_quantity(self):
        return self.__fixed_reorder_quantity

    @fixed_order_quantity.setter
    def fixed_order_quantity(self, order):
        self.__fixed_reorder_quantity = order

    @property
    def economic_order_qty(self):
        return self.__economic_order_qty

    @economic_order_qty.setter
    def economic_order_qty(self, eoq):
        self.__economic_order_qty = eoq

    @property
    def economic_order_variable_cost(self) -> Decimal:
        return self.__economic_order_variable_cost

    @economic_order_variable_cost.setter
    def economic_order_variable_cost(self, eoq_vc):
        self.__economic_order_variable_cost = eoq_vc

    def average_order(self):
        return float(sum(self.__orders.values()) / self.__count_orders)

    def _sum_orders(self) -> int:
        total_orders = 0
        orders_list = []
        for item in self.__orders:
            if isinstance(self.__orders[item], Iterable):
                orders_list = self.__orders[item]
            else:
                orders_list.append(self.__orders[item])

        total_orders = sum([Decimal(item) for item in orders_list])
        return total_orders

    def average_order_row(self) -> Decimal:
        total_orders = 0
        orders_list = []
        for item in self.__orders:
            orders_list = self.__orders[item]
        for item in orders_list:
            total_orders += Decimal(item)
        return Decimal(total_orders / len(orders_list))

    def _revenue(self, orders: dict) -> Decimal:
        orders_list = []
        total_order = 0
        for items in orders:
            orders_list = orders[items]
        if isinstance(orders_list, Iterable):
            for item in orders_list:
                total_order += Decimal(item)
        else:
            for item in orders:
                total_order += orders[item]

        return Decimal(total_order * Decimal(self.__retail_price))

    def _standard_deviation_orders_row(self) -> Decimal:
        getcontext().prec = 12
        deviation = Decimal(0)
        variance = []
        orders_list = []
        for item in self.__orders:
            orders_list = self.__orders[item]
        for item in orders_list:
            variance.append(Decimal(Decimal(item) - Decimal(self.__average_order)))
        for j in variance:
            deviation += Decimal(j) ** Decimal(2)
        deviation /= Decimal(len(variance))
        return Decimal(Decimal(deviation) ** Decimal(0.5))

    # TODO-feature convert lead-time to correct period (data_set period must match lead_time priod if not conversion)
    def _safety_stock(self) -> Decimal:
        return Decimal(self.__z_value) * Decimal(self.__orders_standard_deviation) * Decimal(
            (self.__lead_time ** Decimal(0.5)))

    def _demand_variability(self) -> Decimal:
        return Decimal(Decimal(self.__orders_standard_deviation) / Decimal(self.__average_order))

    def _reorder_level(self) -> Decimal:
        return (Decimal(self.__lead_time ** Decimal(0.5)) * Decimal(self.__average_order)) + Decimal(
            self.__safety_stock)

    # provide the facility to output order quantity as a range if the reorder cost is an estimation
    # one version when holding cost has not been specified and one when it has been
    def _fixed_order_quantity(self) -> Decimal:
        return (2 * Decimal(self.__reorder_cost) * (
            Decimal(self.__average_order) / (
                Decimal(self.__unit_cost) * Decimal(self.__CONST_HOLDING_COST_FACTOR)))) ** Decimal(0.5)

    def _shortage_qty(self):
        if self.__quantity_on_hand < self.__safety_stock:
            return round(
                abs(((self.__reorder_level + (
                    self.__reorder_level - self.__safety_stock)) - self.__quantity_on_hand) + self.__backlog))
        else:
            return 0

    def _excess_qty(self):
        if self.__quantity_on_hand > self.__reorder_level + (self.__reorder_level - self.__safety_stock):
            return round(
                self.__quantity_on_hand - (self.__reorder_level + (self.__reorder_level - self.__safety_stock)), 0)
        else:
            return 0

    @log_this(logging.INFO, message='Called to generate optimised SES forecast.')
    def _generate_optimised_ses_forecast(self):
        try:
            demand = (list(self.__orders.get("demand")))
            orders = [int(i) for i in demand]
            ses_evo_forecast = simple_exponential_smoothing_forecast(demand=orders, smoothing_level_constant=0.5)
            return ses_evo_forecast
        except TypeError as e:
            print('Exponential smoothing forecast (evolutionary model) failed. {}'.format(e))
        except OSError as e:
            print(e)

    @log_this(logging.INFO, message='Called to generate optimised HTCES forecast.')
    def _generate_holts_es_forecast(self):
        demand = (list(self.__orders.get("demand")))
        orders = [int(i) for i in demand]

        htces = holts_trend_corrected_exponential_smoothing_forecast(demand=orders, alpha=0.5, gamma=0.5, optimise=True)

        return htces

    def _summary(self, keywords: list) -> dict:
        pre_build = {'sku': self.__sku_id, 'average_order': '{:.0f}'.format(self.__average_order),
                     'standard_deviation': '{:.0f}'.format(self.__orders_standard_deviation),
                     'safety_stock': '{:.0f}'.format(self.__safety_stock),
                     'demand_variability': '{:.3f}'.format(self.__demand_variability),
                     'reorder_level': '{:.0f}'.format(self.__reorder_level),
                     'reorder_quantity': '{:.0f}'.format(self.__fixed_reorder_quantity),
                     'revenue': '{}'.format(self.__sku_revenue),
                     'economic_order_quantity': '{:.0f}'.format(self.__economic_order_qty),
                     'economic_order_variable_cost': '{:.2f}'.format(self.__economic_order_variable_cost),
                     'ABC_XYZ_Classification': '{0}{1}'.format(self.__abc_classification, self.__xyz_classification),
                     'excess_stock': '{}'.format(self.__excess_stock),
                     'shortages': '{}'.format(self.__shortage_qty),
                     'average_orders': '{}'.format(self.__average_order),
                     'unit_cost': '{}'.format(self.__unit_cost),
                     'quantity_on_hand': '{}'.format(self.__quantity_on_hand),
                     'currency': '{}'.format(self.__currency),
                     'orders': self.__orders,
                     'total_orders': '{}'.format(self.__total_orders),
                     'backlog': '{}'.format(self.__backlog)}
        summary = {}
        for key in keywords:
            summary.update({key: pre_build.get(key)})
        return summary

    def orders_summary(self) -> dict:
        return self._summary(self._summary_keywords)

    def orders_summary_simple(self)->dict:
        return self._summary(
            itertools.chain(self._summary_keywords[:7], self._summary_keywords[9:len(self._summary_keywords)]))

    def __repr__(self):
        representation = "(sku_id: {}, average_order: {:.0f}, standard_deviation: {:.0f}, safety_stock: {:0f}, \n" \
                         "demand_variability: {:.3f}, reorder_level: {:.0f}, reorder_quantity: {:.0f}, " \
                         "revenue: {:.2f}, excess_stock: {}, shortages: {}, unit_cost: {}, quantity_on_hand: {}, " \
                         "currency_code: {}, total_orders: {}, backlog: {})"
        return representation.format(self.__sku_id,
                                     self.__average_order,
                                     self.__orders_standard_deviation,
                                     self.__safety_stock,
                                     self.__demand_variability,
                                     self.__reorder_level,
                                     self.__fixed_reorder_quantity,
                                     self.__sku_revenue,
                                     self.__excess_stock,
                                     self.__shortage_qty,
                                     self.__unit_cost,
                                     self.__quantity_on_hand,
                                     self.__currency,
                                     self.__total_orders,
                                     self.__backlog)

    def __iter__(self):
        for original_order in self.__orders.get("demand"):
            yield original_order

    def __len__(self):
        return len(self.__orders.get("demand"))

    def __del__(self):

        self.__orders = None
        self.__sku_id = None
        self.__lead_time = None
        self.__unit_cost = None
        self.__z_value = None
        self.__count_orders = None
        self.__average_order = None
        self.__orders_standard_deviation = None
        self.__safety_stock = None
        self.__demand_variability = None
        self.__reorder_level = None
        self.__reorder_cost = None
        self.__fixed_reorder_quantity = None



# Copyright (c) 2015-2016, The Authors and Contributors
# <see AUTHORS file>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
# following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
# following disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os

from supplychainpy._helpers._config_file_paths import ABS_FILE_PATH_APPLICATION_CONFIG
from supplychainpy._helpers._pickle_config import deserialise_config


class Config(object):
    pass


class ProdConfig(Config):
    PWD = os.path.abspath(os.curdir)
    DEBUG = False
    conf = deserialise_config(ABS_FILE_PATH_APPLICATION_CONFIG)
    db_path = conf.get('database_path', None)
    if db_path is not None and os.name in ['posix', 'mac']:
        SQLALCHEMY_DATABASE_URI = 'sqlite:///{}/reporting.db'.format(db_path)
    elif os.name == 'nt':
        SQLALCHEMY_DATABASE_URI = 'sqlite:///{}\\reporting.db'.format(db_path)
    SECRET_KEY = '|~G\xde\xa7\x9b\x1aKaZ-\xabk8\x0b\x12\xee)\xe0\xe0\x8b\x0c\xd9\x1d'
    SESSION_PROTECTION = 'strong'
    SQLALCHEMY_ECHO = False
    SQLALCHEMY_TRACK_MODIFICATIONS = True


class DevConfig(Config):
    PWD = os.path.abspath(os.curdir)
    DEBUG = True
    conf = deserialise_config(ABS_FILE_PATH_APPLICATION_CONFIG)
    db_path = conf.get('database_path', None)
    if os.name in ['posix', 'mac']:
        SQLALCHEMY_DATABASE_URI ='sqlite:///{}/reporting.db'.format(db_path)

    elif os.name == 'nt':
        SQLALCHEMY_DATABASE_URI = 'sqlite:///{}\\reporting.db'.format(db_path)
    SECRET_KEY = '|~G\xde\xa7\x9b\x1aKaZ-\xabk8\x0b\x12\xee)\xe0\xe0\x8b\x0c\xd9\x1d'
    SESSION_PROTECTION = 'strong'
    SQLALCHEMY_ECHO = False
    SQLALCHEMY_TRACK_MODIFICATIONS = True

class IntegrationConfig(Config):
    DEBUG = True
    SQLALCHEMY_DATABASE_URI =''
    SECRET_KEY = '|~G\xde\xa7\x9b\x1aKaZ-\xabk8\x0b\x12\xee)\xe0\xe0\x8b\x0c\xd9\x1d'
    SESSION_PROTECTION = 'strong'
    SQLALCHEMY_ECHO = False
    SQLALCHEMY_TRACK_MODIFICATIONS = True


#!/usr/bin/python3
# Copyright (c) 2015-2016, The Authors and Contributors
# <see AUTHORS file>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are 
# permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
# following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
# following disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from decimal import Decimal
from decimal import ROUND_FLOOR
from decimal import getcontext

import numpy as np

from supplychainpy._helpers._enum_formats import PeriodFormats
from supplychainpy.simulations import simulation_window

# assumptions: opening stock in first period is average stock adjusted to period used in monte carlo if different from
# orders analysis. There are no deliveries in the first period (maybe add switch so there always is a delivery in first,
# users choice) period based on inventory rules.


class SetupMonteCarlo:
    """ Create a monte carlo simulation for inventory analysis."""

    _conversion = 1
    _window = {}

    def __init__(self, analysed_orders: list, period: str = PeriodFormats.months.name, period_length: int = 12):
        self._analysed_orders = analysed_orders
        self._normal_random_distribution = self.generate_normal_random_distribution(period_length=period_length)

    @property
    def normal_random_distribution(self):
        return self._normal_random_distribution

    def generate_normal_random_distribution(self, period_length: int) -> list:

        """ Generates the random demand for a given sku.

        For each sku a set of random demands are calculated based on the normal distribution of demand for this product.

        Args:
            period_length (int):    length of window e.g. 12 weeks for a quarter etc.


        Returns:
            list:   A list of randomly generated demand.



        Raises:
            ValueError:

        """
        orders_normal_distribution = {}
        random_orders_generator = []
        final_random_orders_generator = []
        for sku in self._analysed_orders:
            for i in range(0, period_length):
                nrd_orders = np.random.normal(loc=sku.average_orders,
                                              scale=sku.standard_deviation,
                                              size=sku.order_count)

                random_orders_generator.append(abs(np.array(nrd_orders)).tolist())
            orders_normal_distribution[sku.sku_id] = random_orders_generator
            random_orders_generator = []
        final_random_orders_generator.append(orders_normal_distribution)
        return final_random_orders_generator

    # replace implementation with yield generators instead

    def build_window(self, random_normal_demand: list, period_length: int = 0,
                     holding_cost_percentage: Decimal = 0.48,
                     shortage_cost_percentage: Decimal = 0.3) -> dict:

        """ Builds the simulation window for opening stock, demand, closing stock, backlog, holding cost and shortages.

        The window represents a fixed time frame, supplied by the user. The window is then run several times to generate
        probabilities for example the probability of stocking out during the period specified or the probability that the
        final closing stock will be negative. The results will indicate the accuracy of the analytical model used to
        calculate the safety stock etc, based on a few assumptions.

        Args:
            random_normal_demand (list):        A list of random demand normally distributed.
            period_length (int):                length of window e.g. 12 weeks for a quarter etc.
            holding_cost_percentage (Decimal):  The percentage of unit cost to associate with holding cost,
            shortage_cost_percentage (Decimal): The percentage of unit cost to associate with shortage cost,

        Returns:
            dict:   The build_window returns a dictionary of lists for opening_stock, demand, delivery,closing stock,
                    backlog, holding cost, shortage cost.



        Raises:
            ValueError:     The period_length is currently {} and the actual length of the demand is {}.
                            Please make sure that the two values are equal".format(period_length,
                            len(random_normal_demand[0][sku.sku_id]))
        """

        getcontext().prec = 7
        getcontext().rounding = ROUND_FLOOR

        clist = []

        revenue = lambda unit_cost, units_sold,: Decimal(unit_cost) * Decimal(units_sold)

        # lambda functions for calculating the main values in the monte carlo analysis
        closing_stock = lambda opening_stock, orders, deliveries, backlog: Decimal((Decimal(opening_stock)
                                                                                    - Decimal(orders)) + Decimal(
            deliveries)) - Decimal(backlog) if Decimal((Decimal(opening_stock) - Decimal(orders)) +
                                                       Decimal(deliveries)) - Decimal(backlog) > 0 else 0

        backlog = lambda opening_stock, deliveries, demand: Decimal(abs(
            (Decimal(opening_stock + deliveries)) - Decimal(demand))) if \
            Decimal((opening_stock + deliveries)) - Decimal(demand) < 0 else 0

        holding_cost = lambda cls_stock, unit_cost: cls_stock * (
            Decimal(unit_cost) * Decimal(holding_cost_percentage)) if cls_stock > 0 else 0

        shortages = lambda opening_stock, orders, deliveries: abs((Decimal(opening_stock) - Decimal(orders)) +
                                                                  Decimal(deliveries)) if \
            ((Decimal(opening_stock) - Decimal(orders)) + Decimal(deliveries)) < 0 else 0

        # shortage cost as a percentage of unit cost. Increase to percentage to have a bigger affect change more
        shortage_cost = lambda cls_stock, unit_cost: cls_stock * (
            Decimal(unit_cost) * Decimal(shortage_cost_percentage)) if int(cls_stock) > 0 else 0

        raise_po = lambda reorder_lvl, cls_stock: True if cls_stock <= reorder_lvl else False

        po_qty = lambda eoq, reorder_lvl, backlog, cls_stock: Decimal(eoq) + Decimal(backlog) + Decimal(
            (Decimal(reorder_lvl) - Decimal(cls_stock))) if Decimal(eoq) + Decimal(backlog) + Decimal(
            (Decimal(reorder_lvl) - Decimal(cls_stock))) > 0 else 0

        # calculate period to receive po and quantity to receive

        for sku in self._analysed_orders:
            if period_length != len(random_normal_demand[0][sku.sku_id]):
                raise ValueError("The period_length is currently {} and the actual length of the demand is {}. "
                                 "Please make sure that the two values are equal".format(period_length,
                                                                                         len(random_normal_demand[0][
                                                                                                 sku.sku_id])))

        sim_frame_collection = []
        index_item = 1
        for sku in self._analysed_orders:

            period = 1
            order_receipt_index = {}
            final_stock = 0
            sim_window_collection = {}
            previous_backlog = Decimal('0')
            order_receipt_index = {}
            # create the sim_window for each sku, suing the random normal demand generated
            for i in range(0, period_length):

                po_qty_raised = 0

                # instantiate sim_window
                sim_window = simulation_window.MonteCarloWindow

                # add sku_id
                sim_window.sku_id = sku.sku_id

                # add closing stock
                previous_closing_stock = final_stock

                # mark sim_window.position or period in analysis
                sim_window.position = period

                # add average orders to opening_stock if first period else add closing stock
                if sim_window.position == 1:
                    sim_window.opening_stock = (sku.reorder_level - Decimal(sku.safety_stock)) + Decimal(sku.safety_stock) #calculated ltd until put into analyse orders
                else:
                    sim_window.opening_stock = previous_closing_stock

                # add random demand
                demand = random_normal_demand[0][sku.sku_id][i][0]
                sim_window.demand = demand

                #
                if sim_window.position in order_receipt_index.keys():
                    sim_window.purchase_order_receipt_qty = order_receipt_index[sim_window.position]
                    sim_window.po_number_received = 'PO {:.0f}{}'.format(sim_window.position, sim_window.index)
                    del order_receipt_index[sim_window.position]
                else:
                    sim_window.purchase_order_receipt_qty = 0
                    sim_window.po_number_received = ''

                sim_window.index = index_item

                sim_window.backlog = backlog(opening_stock=sim_window.opening_stock,
                                             deliveries=sim_window.purchase_order_receipt_qty,
                                             demand=demand) + previous_backlog
                sim_window.closing_stock = closing_stock(opening_stock=sim_window.opening_stock,
                                                         orders=demand,
                                                         deliveries=sim_window.purchase_order_receipt_qty,
                                                         backlog=sim_window.backlog)

                sim_window.holding_cost = holding_cost(sim_window.closing_stock, sku.unit_cost)

                sim_window.shortage_units = shortages(opening_stock=sim_window.opening_stock,
                                                      orders=demand,
                                                      deliveries=sim_window.purchase_order_receipt_qty)

                sim_window.shortage_cost = shortage_cost(cls_stock=(sim_window.backlog - previous_backlog),
                                                         unit_cost=sku.unit_cost)

                sim_window.po_raised_flag = raise_po(reorder_lvl=sku.reorder_level, cls_stock=sim_window.closing_stock)

                po_receipt_period = period + sku.lead_time

                po_qty_raised = po_qty(eoq=sku.economic_order_qty,
                                       reorder_lvl=sku.reorder_level,
                                       backlog=sim_window.backlog,
                                       cls_stock=sim_window.closing_stock)

                if po_qty_raised > 0:
                    order_receipt_index.update({po_receipt_period: po_qty_raised})
                    sim_window.purchase_order_raised_qty = order_receipt_index.get(po_receipt_period)
                else:
                    sim_window.purchase_order_raised_qty = 0

                sim_window.po_number_raised = ''

                if int(sim_window.purchase_order_raised_qty) > 0:
                    sim_window.po_number_raised = 'PO {:.0f}{}'.format(po_receipt_period, sim_window.index)
                    del po_receipt_period

                final_stock = sim_window.closing_stock

                if int(sim_window.closing_stock) == 0:
                    previous_backlog += sim_window.backlog
                else:
                    previous_backlog = 0

                units_sold = self._units_sold(backlog=sim_window.backlog, opening_stock=sim_window.opening_stock,
                                              delivery=sim_window.purchase_order_receipt_qty, demand=sim_window.demand)
                sim_window.sold = units_sold
                sim_window.revenue = Decimal(revenue(sku.unit_cost, units_sold))
                yield sim_window

                del sim_window
                del po_qty_raised
                period += 1

            index_item += 1

    def _units_sold(self, backlog, opening_stock, delivery, demand):

        # check if opening_stock + closing_stock = 0
        if int(opening_stock) + int(delivery) == 0:
            sold = 0.00
        else:
            sold = (opening_stock + delivery) - Decimal(demand) - Decimal(backlog)

        if sold < 0:
            units_sold = Decimal(demand) + Decimal(backlog) - Decimal(abs(sold))
        else:
            units_sold = Decimal(sold)

        return units_sold



def sum_squared_errors_i1(squared_error: list, alpha: float, gamma: float):
        cdef double sse = 0

        for sq_e in squared_error:
            for i in sq_e:
                if i['alpha'] == alpha and i['gamma'] == gamma:
                    sse += i["squared_error"]

        return {(alpha, gamma): sse}

def sum_squared_errors_i2(squared_error: list, smoothing_parameter: float):
    cdef double sse = 0

    for sq_e in squared_error:
        for i in sq_e:
            if i['alpha'] == smoothing_parameter:
                sse += i["squared_error"]

    return {smoothing_parameter: sse}

def sum_squared_errors_i0(squared_error: list, smoothing_parameter: float):
    cdef double sse = 0
    for sq_e in squared_error:
        if sq_e['alpha'] == smoothing_parameter:
            sse += sq_e["squared_error"]
    return {smoothing_parameter: sse}


# Copyright (c) 2015-2016, The Authors and Contributors
# <see AUTHORS file>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
# following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
# following disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import datetime

from flask import Flask

from supplychainpy._helpers._db_connection import database_connection_uri
from supplychainpy.reporting.blueprints.models import SimulationSummary, Simulation, MasterSkuList
from supplychainpy.reporting.config.settings import ProdConfig
from supplychainpy.reporting.extensions import db


def _retrieve_sku_id(db, sim):
    return db.session.query(MasterSkuList.id).filter(MasterSkuList.sku_id == sim.get('sku_id', 0)).first()


def select_last_simulation():
    app = Flask(__name__)
    app.config.from_object(ProdConfig)
    db.init_app(app)
    app.config['SQLALCHEMY_DATABASE_URI'] = database_connection_uri(retrieve='retrieve')
    with app.app_context():
        simulation_sub = db.session.query(db.func.max(Simulation.date))
        simulation_id = db.session.query(Simulation).filter(Simulation.date == simulation_sub).first()
        return db.session.query(SimulationSummary).filter(SimulationSummary.sim_id == simulation_id.id).all()

def store_simulation(sim_summary: dict):
    app = Flask(__name__)
    app.config.from_object(ProdConfig)
    db.init_app(app)
    app.config['SQLALCHEMY_DATABASE_URI'] = database_connection_uri(retrieve='retrieve')
    with app.app_context():
        simulation = Simulation()
        simulation.date = datetime.datetime.now()
        db.session.add(simulation)
        db.session.commit()
        simulation_sub = db.session.query(db.func.max(Simulation.date))
        simulation_id = db.session.query(Simulation).filter(Simulation.date == simulation_sub).first()
        for sim in sim_summary:
            master_sku_id = _retrieve_sku_id(db, sim)
            simulation_summary = SimulationSummary()
            simulation_summary.sku_id = master_sku_id.id
            simulation_summary.sim_id = simulation_id.id
            simulation_summary.average_backlog = sim.get('average_backlog', 0)
            simulation_summary.variance_opening_stock = sim.get('variance_opening_stock', 0)
            simulation_summary.minimum_opening_stock = sim.get('minimum_opening_stock', 0)
            simulation_summary.maximum_closing_stock = sim.get('maximum_closing_stock', 0)
            simulation_summary.maximum_quantity_sold = sim.get('maximum_quantity_sold', 0)
            simulation_summary.maximum_backlog = sim.get('maximum_backlog', 0)
            simulation_summary.average_backlog = sim.get('average_backlog', 0)
            simulation_summary.minimum_closing_stock = sim.get('minimum_closing_stock', 0)
            simulation_summary.minimum_backlog = sim.get('minimum_backlog', 0)
            simulation_summary.minimum_quantity_sold = sim.get('minimum_quantity_sold', 0)
            simulation_summary.average_quantity_sold = sim.get('average_quantity_sold', 0)
            simulation_summary.maximum_opening_stock = sim.get('maximum_opening_stock', 0)
            simulation_summary.standard_deviation_quantity_sold = sim.get('standard_deviation_quantity_sold', 0)
            simulation_summary.standard_deviation_closing_stock = sim.get('standard_deviation_closing_stock', 0)
            simulation_summary.average_closing_stock = sim.get('average_closing_stock', 0)
            simulation_summary.standard_deviation_backlog = sim.get('standard_deviation_backlog', 0)
            simulation_summary.average_shortage_units = sim.get('average_shortage_units', 0)
            db.session.add(simulation_summary)
            db.session.commit()



# Copyright (c) 2015-2016, The Authors and Contributors
# <see AUTHORS file>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
# following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
# following disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import logging

from supplychainpy.bi._analytical_heirachy_process import _PairwiseComparison

log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())


def analytical_hierarchy_process(criteria: tuple, criteria_scores: list, options: tuple, option_scores: dict,
                                 quantitative_criteria: tuple = None, **kwargs):
    """ Compute an analytical hierarchy for alternative choices based on relative weights and priorities of categories.

    Args:
        criteria (tuple):                List the criteria to base decision between alternative choices.
        criteria_scores (list):          List the scores for the criteria as a list of tuples, the reciprocals are
                                         auto-calculated.
        options (tuple):                 The alternative otpions to decide between.
        option_scores (dict):            A dict of tuples scoring the alternatives. The tuple order must mirror the
                                         'alternative' argument tuple.
        quantitative_criteria (tuple):   List the quantitative criteria e.g. 'fuel economy', that use representative
                                         values and not scores
        **kwargs:
            item_cost (dict):   A dict of alternatives and their corresponding costs. Only necessary if seeking to use cost benefit ratios for further discretion.

    Returns:
        dict:   Summary of analytical hierarchy and cost benefit analysis if kwargs 'item_cost' used.

    Examples:

    >>> lorry_cost = {'scania': 68000,'iveco': 79000,'volvo': 59000,'navistar': 66000}
    >>> criteria = ('style', 'reliability', 'fuel_economy')
    >>> criteria_scores = [(1, 1 / 2, 3), (0, 1, 4), (0, 0, 1)]
    >>> options = ('scania', 'iveco', 'volvo', 'navistar')
    >>> options_scores ={'reliability': [(1, 2, 5, 1), (1 / 2, 1, 3, 2), (1 / 5, 1 / 3, 1, 1 / 4), (1, 1 / 2, 4, 1)],
    ...     'style': [(1, 1 / 4, 4, 1 / 6), (4, 1, 4, 1 / 4),(1 / 4, 1 / 4, 1, 1 / 5), (6, 4, 5, 1)],
    ...     'fuel_economy': (62, 55, 56, 56)}
    >>> lorry_decision = analytical_hierarchy_process(criteria= criteria, criteria_scores=criteria_scores,
    ...     options= options, option_scores=options_scores, quantitative_criteria=('fuel_economy',),
    ...     item_cost = lorry_cost)

    """
    ahp = _PairwiseComparison(criteria=criteria,
                              criteria_scores=criteria_scores,
                              options=options,
                              option_scores=option_scores,
                              quantitative_criteria=quantitative_criteria)
    if kwargs:
        ahp_cvb = ahp.cost_benefit_summary(ahp_summary=ahp.summary(), item_cost=kwargs.get('item_cost'))
        log.debug("AHP completed")
        return {'analytical_hierarchy': ahp.summary(), 'cost_benefit_ratios': ahp_cvb}

    return ahp.summary()



from behave import given, when, then
from supplychainpy.bot._dash_states import DashStates
from supplychainpy.bot.dash import ChatBot


@given("I initiate chat with bot with a greeting")
def step_impl(context):
    """
    Args:
        context (behave.runner.Context):
    """
    bot = ChatBot()
    assert isinstance(bot, ChatBot)


@when("I say hello to the bot")
def step_impl(context):
    """
    Args:
        context (behave.runner.Context):
    """


    bot = ChatBot()
    context.response = bot.chat_machine(message='Hello')
    assert context.response


@then("It should respond")
def step_impl(context):
    """
    Args:
        context (behave.runner.Context):
    """
    dash = DashStates()
    assert context.response[0][0] in dash.salutations





# Copyright (c) 2015-2016, The Authors and Contributors
# <see AUTHORS file>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
# following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
# following disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import heapq
from copy import deepcopy
from decimal import Decimal

from supplychainpy._helpers._codes import Currency
from supplychainpy._helpers._pickle_config import serialise_config
from supplychainpy.inventory.analyse_uncertain_demand import UncertainDemand
from supplychainpy.inventory.summarise import Inventory
from supplychainpy.sample_data.config import ABS_FILE_PATH


class ResponseBorg:
    """ Borg class making  class attributes global """

    _shared_response = {}  # global attribute dictionary

    def __init__(self):
        self.__dict__ = self._shared_response


class ResponseSingleton(ResponseBorg):
    """ Singleton class using global dict in ResponseBorg"""

    def __init__(self, **kwargs):
        ResponseBorg.__init__(self)
        self._shared_response.update(kwargs)

    @property
    def shared_response(self) -> dict:
        return self._shared_response

    def __str__(self):
        # returns the attribute for printing
        return str(self._shared_response)


class SKUStates:
    """Contains states required to generate recommendations for individual SKUs"""
    _END_STATE = 'recommendation'
    _EMPTY = 'EMPTY'
    _TRANSITION_STATES = {
        'EXCESS_RANK_STATE': 'excess_rank',
        'SHORTAGE_RANK_STATE': 'shortage_rank',
        'INVENTORY_TURNS_STATE': 'inventory_turns',
        'CLASSIFICATION_STATE': 'classification',
        'TRAFFIC_LIGHT_STATE': 'traffic_light',
        'FORECAST_STATE': 'forecast',
        'RECOMMENDATION_STATE': 'recommendation'
    }

    def __init__(self, analysed_orders: UncertainDemand, forecast: dict = None):
        self._analysed_orders = analysed_orders
        self._summarised_inventory = Inventory(analysed_orders)
        self._htces_forecast = forecast
        self._summary = {}
        self._compiled_response = ResponseSingleton()
        self._currency_symbol = self._retrieve_currency()

    @property
    def currency(self):
        return self._currency_symbol

    @property
    def analysed_orders(self) -> UncertainDemand:
        return self._analysed_orders

    @property
    def summarised_inventory(self) -> Inventory:
        return self._summarised_inventory

    @property
    def htces_forecast(self) -> dict:
        return self._htces_forecast

    @property
    def compiled_response(self) -> ResponseSingleton:
        return self._compiled_response

    @property
    def summary(self) -> dict:
        return self._summary

    @summary.setter
    def summary(self, value):
        self._summary = value

    def _retrieve_currency(self) -> str:
        currency_code = ''
        for i in self.analysed_orders:
            currency_code = i.currency
            break
        symbol = Currency(code=currency_code)
        return symbol.retrieve_symbol()

    def _setup_summary(self, sku: str):
        """Prepares summary for each sku.

        Args:
            sku (str):  SKU unique identification

        """
        self.summary = [description for description in
                        self._summarised_inventory.describe_sku(sku)][0]

    def initialise_machine(self, sku: str) -> tuple:
        """Initialises state machine.

        Args:
            sku (str): SKU unique identification

        Returns:
            tuple:  New State and sku unique identification

        """
        self._setup_summary(sku=sku)
        excess_units = int(self._summary.get('excess_units'))
        shortage_units = int(self._summary.get('shortage_units'))
        if excess_units > 0:
            state = self._TRANSITION_STATES.get('EXCESS_RANK_STATE', self._END_STATE)
        elif shortage_units > 0:
            state = self._TRANSITION_STATES.get('SHORTAGE_RANK_STATE', self._END_STATE)
        else:
            state = self._TRANSITION_STATES.get('INVENTORY_TURNS_STATE', self._END_STATE)
        return state, sku

    def append_response(self, response: str, sku: str):
        """ Appends each response to the dict in the singleton.

        Args:
            response:
            sku:

        Returns:

        """

        resp = self.compiled_response.shared_response.get(sku, self._EMPTY)
        if self._EMPTY != resp:
            response = resp + response
            self.compiled_response.shared_response.update(**{'{}'.format(sku): response})
        else:
            self.compiled_response.shared_response.update(**{'{}'.format(sku): response})

    def excess_rank(self, sku: str) -> tuple:
        """Excess State.

        Args:
            sku (str):    SKU unique identification number.

        Returns:
            str:    New state and SKU ID

        """
        excess_rank = int(self._summary.get('excess_rank', 11))

        if excess_rank <= 10:
            response = '{} is one of the top 10 overstocked SKUs in your inventory profile, ranked {} of ' \
                       '10 overstocked SKUs. '.format(sku, sku, excess_rank)

            self.append_response(response=response, sku=sku)
            sku_classification = self._summary.get('classification', 'XX')
            if sku_classification == 'AX':
                quantity_on_hand = self._summary.get('quantity_on_hand')
                reorder_quantity = self._summary.get('reorder_quantity')
                response = 'The {} inventory classification indicates a stable demand profile. ' \
                           'SKU {} is in the 20% of SKU\'s that contribute 80% yearly revenue. ' \
                           'Unless {} product is approaching end of life (EOL)' \
                           ', there is likely little need to panic. ' \
                           'Holding purchase orders for this SKU will reduce ' \
                           'the excess. Review consumption of this SKU until the quantity on hand, ' \
                           'currently {}, is close to or below the reorder level of {}. '.format(sku_classification,
                                                                                                 sku, sku,
                                                                                                 quantity_on_hand,
                                                                                                 reorder_quantity)
                self.append_response(response=response, sku=sku)

        state = self._TRANSITION_STATES.get('INVENTORY_TURNS_STATE', self._END_STATE)
        return state, sku

    def shortage_rank(self, sku: str) -> tuple:
        """Shortage rank state

        Args:
            sku (str):    SKU unique identification.

        Returns:
            tuple:        New state and SKU ID

        """
        shortage_rank = int(self._summary.get('shortage_rank', 11))
        if shortage_rank <= 10:
            response = '{} is one of the top 10 understocked SKUs ' \
                       'in your inventory profile. {} is currently ranked {} of ' \
                       '10 understocked SKUs. '.format(sku, sku, shortage_rank)
            self.append_response(response=response, sku=sku)

        state = self._TRANSITION_STATES.get('INVENTORY_TURNS_STATE', self._END_STATE)
        return state, sku

    def inventory_turns(self, sku: str) -> tuple:
        """Inventory turns state.

        Args:
            sku:    SKU unique identification.

        Returns:
            tuple:  New state and SKU ID

        """
        inventory_turns = float(self._summary.get('inventory_turns'))
        excess_rank = int(self._summary.get('excess_rank', 11))
        if inventory_turns <= 2.00 and excess_rank <= 10:
            response = '{} has a very low rolling inventory turn rate at {:.2f}. {} may be a greater cause for ' \
                       'concern considering it is also in the top 10 overstocked ' \
                       'SKUs in the inventory profile. '.format(sku, float(self._summary.get('inventory_turns')), sku)

            self.append_response(response=response, sku=sku)
        elif inventory_turns <= 2.00:
            response = '{} has a very low rolling inventory turn rate at {:.2f}. '.format(
                sku, float(self._summary.get('inventory_turns')))
            self.append_response(response=response, sku=sku)

        state = self._TRANSITION_STATES.get('CLASSIFICATION_STATE', self._END_STATE)
        return state, sku

    def classification(self, sku: str) -> tuple:
        """Classification state.

        Args:
            sku (str):    SKU ID number

        Returns:
           tuple:     New state and SKU ID

        """

        classification = self._summary.get('classification')
        unit_cost_rank = self._summary.get('unit_cost_rank')
        excess_rank = int(self._summary.get('excess_rank', 11))

        if classification in ('AZ', 'BZ', 'CZ'):
            response = 'The {} classification indicates that {} has volatile demand, ' \
                       'making it difficult to predict' \
                       ' the consumption of the excess. '.format(self._summary.get('classification'), sku)
            self.append_response(response=response, sku=sku)

        if classification == 'CZ' and unit_cost_rank < 10 and excess_rank <= 10:
            response = '{} is a slow moving volatile SKU. A unit cost of {} makes it also one of the more costly' \
                       ' in the inventory profile.  ' \
                       'It may be prudent to make financial provisions for this product, ' \
                       'discount them or bundle them with a complimentary popular product ' \
                       'in a promotion. '.format(sku, self._summary.get('unit_cost'))
            self.append_response(response=response, sku=sku)

        if classification in ('AX', 'BX'):
            response = 'The {} classification, indicates that {} has a stable demand profile and ' \
                       'contributes significantly to revenue. ' \
                       'Corrective action should be taken to bring the quantity on hand (QOH) up from {} to {}. '
            self.append_response(response=response, sku=sku)

        state = self._TRANSITION_STATES.get('TRAFFIC_LIGHT_STATE', self._END_STATE)
        return state, sku

    def traffic_light(self, sku: str) -> tuple:
        """ Traffic light state.

        Args:
            sku (str):    SKU ID number

        Returns:
            tuple:  New state and SKU ID

        """
        traffic_light = self._summary.get('inventory_traffic_light')
        excess_units = int(self._summary.get('excess_units', 11))
        classification = self._summary.get('classification')
        shortage_units = int(self._summary.get('shortage_units', 11))

        if traffic_light == 'white':
            response = 'The QOH is less than 75% of safety stock, implying a substantial depletion of ' \
                       'buffer stock. Take into consideration that it is acceptable ' \
                       'to dip into safety stock approximately 50% of the time. However, this situation ' \
                       'poses a threat for servicing future demand. This situation may be acceptable if this product ' \
                       'is \'end of life\' and the goal is to drive down stock. Corrective action is ' \
                       'required if it is not currently a policy to reduce the QOH for {}, and the ' \
                       'receipt of a purchase order is not imminent. '.format(sku)
            self.append_response(response=response, sku=sku)
        elif traffic_light == 'red':
            response = 'The QOH is less 50% of the recommended safety stock. ' \
                       'Take into consideration that it is acceptable ' \
                       'to dip into safety stock approximately 50% of the time, consuming ' \
                       'approximately 50% of its value.' \
                       'Therefore, in this situation there may be little imminent threat to the service level. ' \
                       'Checking that a purchase order has been placed may be prudent. '.format(sku)
            self.append_response(response=response, sku=sku)
        elif traffic_light == 'amber':
            response = 'The QOH is less than the reorder level but has yet to hit safety stock. ' \
                       'There is little to worry about at this point. The periodic review of inventory ' \
                       'and communication of any upcoming deals or promotions ' \
                       'that may significantly impact the demand profile should be a focus.'
            self.append_response(response=response, sku=sku)
        elif traffic_light == 'green' and excess_units == 0 and classification not in ('CZ', 'CY', 'BZ', 'AZ') \
                and shortage_units == 0:

            response = 'Congratulations the QOH is within optimal boundaries.  ' \
                       'There is little to worry about at this point. The periodic review of stock ' \
                       'and communication of any upcoming deals or promotions that may significantly ' \
                       'impact the demand profile should be a focus. '
            self.append_response(response=response, sku=sku)
        state = self._TRANSITION_STATES.get('FORECAST_STATE', self._END_STATE)
        return state, sku

    def forecast(self, sku: str) -> tuple:
        """Forecast state.

        Args:
            sku (str):    SKU unique identification number

        Returns:
            tuple:

        """
        # if the demand shows a linear trend then check if the forecast is less than the excess units.
        state = self._TRANSITION_STATES.get('RECOMMENDATION_STATE', self._END_STATE)
        try:
            if self._htces_forecast.get(sku)['statistics'].get('trend'):
                if self._htces_forecast.get(sku)['forecast'][0] < int(self._summary.get('excess_units')):
                    response = 'It is unlikely {} will fair better next month as the most optimistic forecast is '
                    self.append_response(response=response, sku=sku)
            if self._htces_forecast.get(sku)['statistics'].get('trend') and int(self._summary.get('excess_units')) > 0 \
                    and self._summary.get('classification') in ('AX', 'AY', 'BX', 'BY', 'CX') \
                    and self._htces_forecast.get(sku)['forecast'][0] > int(self._summary.get('excess_units')):
                response = 'The current excess can be reduced by reducing purchase orders and allow the ' \
                           'forecasted demand to be catered for from stock. '
                self.append_response(response=response, sku=sku)
            state = self._TRANSITION_STATES.get('RECOMMENDATION_STATE', self._END_STATE)
            serialise_config(configuration=self._compiled_response.shared_response,
                             file_path=ABS_FILE_PATH['RECOMMENDATION_PICKLE'])
        except TypeError as err:
            pass
        return state, sku


class ProfileStates(SKUStates):
    _TRANSITION_STATES = {
        'EXCESS_STATE': 'excess',
        'SHORTAGE_STATE': 'shortage',
        'CLASSIFICATION_STATE': 'classification',
        'REVENUE_STATE': 'revenue',
        'RECOMMENDATION_STATE': 'recommendation',
        'INVENTORY_STATE': 'inventory'
    }

    def __init__(self, analysed_orders: UncertainDemand, forecast: dict = None):
        super(ProfileStates, self).__init__(analysed_orders=analysed_orders, forecast=forecast)
        self._total_revenue = self._total_revenue()

    def initialise_machine(self, sku: str = None):
        state = self._TRANSITION_STATES.get('REVENUE_STATE', self._END_STATE)
        return state

    @property
    def total_revenue(self):
        return self._total_revenue

    def _total_revenue(self) -> Decimal:
        """ Calculates the value for total revenue.
        Returns:
            Decimal: The total revenue of the skus in the inventory profile.

        """
        total_revenue_all_skus = Decimal(0)
        for i in self.analysed_orders:
            total_revenue_all_skus += i.revenue
        return total_revenue_all_skus

    def append_response(self, response: str, sku: str):
        """ Appends each response to the dict in the singleton.

        Args:
            response (str):    Recommendation for transition state
            sku (str):         SKU unique identification number

        """

        resp = self.compiled_response.shared_response.get("profile", "")
        response = resp + response
        self.compiled_response.shared_response.update(**{'{}'.format("profile"): response})

    def _percentage_of_revenue(self, revenue):
        return (revenue / self.total_revenue) ** 100

    def _rank_summary(self, attribute: str, count: int) -> list:
        """ Summarised rank

        Args:
            attribute (str):    SKU summary attribute to rank.
            count (int):        Number of items to rank.

        Returns:

        """
        return [item for item in self.summarised_inventory.rank_summary(attribute=attribute,
                                                                        count=count, reverse=True)]

    def revenue(self, sku: str = None) -> str:
        """

        Args:
            sku (str):  SKU unique identification.

        Returns:
            str:        New state.

        """
        top_10_revenue = self._rank_summary(attribute='revenue', count=10)
        top_10_revenue_skus = [summary.get('sku') for summary in top_10_revenue]
        top_10_revenue_total = sum([Decimal(summary.get('revenue')) for summary in top_10_revenue])

        response = 'SKUs {SKUs} are the top ten contributors to revenue, generating ' \
                   '{currency_symbol}{revenue:,.2f} of revenue. '.format(**{'SKUs': ", ".join(top_10_revenue_skus)},
                                                                        **{'currency_symbol': self.currency},
                                                                        **{'revenue': top_10_revenue_total}, )

        self.append_response(response=response, sku=sku)
        state = self._TRANSITION_STATES.get('CLASSIFICATION_STATE', self._END_STATE)
        return state

    def classification(self, sku: str = None) -> str:
        """ Classification state.

        Args:
            sku (str):   SKU unique identification.

        Returns:
            str:        New state.

        """
        inventory_class = ('AX', 'AY', 'AZ', 'BX', 'BY', 'BZ', 'CX', 'CY', 'CZ')
        classification_revenue = {}

        for i in inventory_class:
            for classifications in self.summarised_inventory.abc_xyz_summary(classification=(i,),
                                                                             category=('revenue',)):
                classification_revenue.update(deepcopy({i: classifications.get(i)['revenue']}))

        # response = 'The revenue by inventory classification can be broken down as: {breakdown}'.format(
        # **{'breakdown': ', '.join(['{}: {currency_symbol}{val:,.2f}'.format(key, **{'val':value},
        # **{'currency_symbol': self.currency}) for (key, value) in classification_revenue.items()])})

        response = 'The percentage contribution to revenue by classification of SKU, can be broken down as follows: ' \
                   '{breakdown} '.format(**{'breakdown': ', '.join(
            ['{}: {val:,.2f}%'.format(key, **{'val': (Decimal(value) / self.total_revenue) * 100}) for (key, value)
             in classification_revenue.items()])})
        self.append_response(response=response, sku=sku)
        top_revenue_classification = max(classification_revenue, key=lambda key: classification_revenue[key])

        if top_revenue_classification in ('AZ', 'BZ', 'CZ'):
            response = '{top} contributing the most to the inventory revenue may'.format(
                **{'top': top_revenue_classification})
            self.append_response(response=response, sku=sku)

        state = self._TRANSITION_STATES.get('EXCESS_STATE', self._END_STATE)
        return state

    def excess(self, sku: str = None) -> str:
        """ Excess state

        Args:
            sku (str):      SKU unique identification.

        Returns:
            str:            New state.

        """
        top_10_excess = [item for item in self.summarised_inventory.rank_summary(attribute='excess_stock',
                                                                                 count=10, reverse=True)]
        excess_skus = [i.get('sku') for i in top_10_excess]
        total_excess_cost = sum([Decimal(description.get('excess_cost')) for description in
                                 self.summarised_inventory.describe_sku(*excess_skus)])
        sku_inventory_turns = {excess_skus[index]: description.get('inventory_turns') for index, description in
                               enumerate(self.summarised_inventory.describe_sku(*excess_skus))}

        response = 'Focus on reducing the excess for the following {sku}. These SKU account for {cost} of excess cost. '.format(
            **{'sku': ", ".join(excess_skus)}, **{'cost': total_excess_cost})

        self.append_response(response=response, sku=sku)
        state = self._TRANSITION_STATES.get('SHORTAGE_STATE', self._END_STATE)
        return state

    def shortage(self, sku: str = None) -> str:

        top_10_shortage = [item for item in self.summarised_inventory.rank_summary(attribute='shortages',
                                                                                   count=10, reverse=True)]

        shortage_skus = [i.get('sku') for i in top_10_shortage]

        total_shortages = sum([Decimal(description.get('shortage_cost')) for description in
                               self.summarised_inventory.describe_sku(*shortage_skus)])

        top_10_revenue = self._rank_summary(attribute='revenue', count=10)
        top_10_revenue_skus = [summary.get('sku') for summary in top_10_revenue]

        shortage_skus_top_revenue = [i.get('sku') for i in top_10_shortage if i.get('sku') in top_10_revenue_skus]

        shortage_skus_top_revenue_total = sum([Decimal(description.get('revenue')) for description in
                                               self.summarised_inventory.describe_sku(*top_10_revenue_skus)])

        response = 'Consider increasing the stock holding of the following SKU: {sku}. ' \
                   'The total cost of covering the shortages is {cost}. ' \
                   'Give more consideration to SKUs {rev_sku} as they are particularly important, ' \
                   'they are in the top 10 contributors to yearly revenue, they contribute {contr_rev:.2f} which is ' \
                   '{perc: .2f}% of total revenue.'.format(
            **{'sku': ', '.join(shortage_skus)}, **{'cost': total_shortages},
            **{'rev_sku': ', '.join(shortage_skus_top_revenue)},
            **{'contr_rev': shortage_skus_top_revenue_total},
            **{'perc': (shortage_skus_top_revenue_total / self.total_revenue) * 100})

        self.append_response(response=response, sku=sku)

        state = self._TRANSITION_STATES.get('INVENTORY_STATE', self._END_STATE)
        return state

    def inventory_turns(self, sku: str = None) -> str:
        """ Inventory turns state.

        Args:
            sku (str):   SKU unique identification.

        Returns:
            str:         New state
        """
        skus = [i.sku_id for i in self.analysed_orders]

        sku_inventory_turns = {skus[index]: description.get('inventory_turns') for index, description in
                               enumerate(self.summarised_inventory.describe_sku(*skus))}

        sku_unit_cost = {skus[index]: description.get('unit_cost') for index, description in
                         enumerate(self.summarised_inventory.describe_sku(*skus))}

        slow_turners = heapq.nsmallest(5, sku_inventory_turns, key=lambda s: sku_inventory_turns[s])
        expensive_stock = heapq.nlargest(5, sku_unit_cost, key=lambda s: sku_inventory_turns[s])

        slow_turning_expensive_stock = [sku for sku in slow_turners if sku in expensive_stock]

        if len(slow_turning_expensive_stock) > 0:
            response = '{sku} experience the slowest inventory turns in the inventory profile, while simultaneously ' \
                       'being the most costly items.'
            self.append_response(response=response, sku=sku)

        serialise_config(configuration=self._compiled_response.shared_response,
                         file_path=ABS_FILE_PATH['PROFILE_PICKLE'])
        state = self._TRANSITION_STATES.get('RECOMMENDATION_STATE', self._END_STATE)
        return state



# Copyright (c) 2015-2016, The Authors and Contributors
# <see AUTHORS file>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
# following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
# following disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import pickle
import re
import logging
import time

from functools import wraps

import sys

import io

from supplychainpy._helpers._config_file_paths import ABS_FILE_PATH_DASH

UNKNOWN = 'UNKNOWN'

def quiet_console(func):
    @wraps(func)
    def wrapper(**kwargs):
        stdout_cache = sys.stdout
        sys.stdout = io.StringIO()
        result = func( **kwargs)
        sys.stdout = stdout_cache
        return result
    return wrapper

def log_this(level, name=None, message=None):
    """Logs and reports the execution time of methods and functions"""

    def time_this(func):
        logname = name if name else func.__module__
        log = logging.getLogger(__name__)
        log.addHandler(logging.NullHandler())
        log_msg = message if message else ''

        @wraps(func)
        def wrapper(*args, **kwargs):
            start = time.perf_counter()
            result = func(*args, **kwargs)
            end = time.perf_counter()
            log.log(level, 'FUNCTION: {}.{} EXECUTION TIME: {} MESSAGE: {}'
                    .format(logname, func.__name__, end - start, log_msg))
            return result

        return wrapper

    return time_this


def preprocess_text(func):
    """ Cleans up messy pronouns

    Args:
        func:

    Returns:

    """

    @wraps(func)
    def wrapper(*args):
        i_regex = re.compile('[iI][ ]')
        iam_regex = re.compile('[iI][\'][mM]')
        im_regex = re.compile('[iI][mM]')
        cleaned = []
        words = args[1].split(' ')
        for word in words:
            if i_regex.match(word):
                cleaned.append('I')
            elif iam_regex.match(word):
                cleaned.append("I am")
            elif im_regex.match(word):
                cleaned.append("I am")
            else:
                cleaned.append(word)
        message = ' '.join(cleaned)
        args = (args[0], message)
        return func(*args)

    return wrapper


def strip_punctuation(func):
    @wraps(func)
    def wrapper(*args):
        word = str(args[1])
        args = (args[0], word.strip('.'))
        return func(*args)

    return wrapper


def pickle_response(func):
    @wraps(func)
    def wrapper(*args):
        message = func(*args)
        state_map = {'last_utterance': message}
        with open(ABS_FILE_PATH_DASH, 'wb') as f:
            pickle.dump(state_map, f)
        return message

    return wrapper


def keyword_sniffer(func):
    @wraps(func)
    def wrapper(**kwargs):
        temp_kwargs = {}
        temp_kwargs.update({'file_path': kwargs.get('file_path', UNKNOWN),
                            'df': kwargs.get('df', UNKNOWN),
                            'raw_data': kwargs.get('raw_data', UNKNOWN),
                            'file_type': kwargs.get('file_type', UNKNOWN),
                            'retail_price': kwargs.get('raw_data', UNKNOWN),
                            'quantity_on_hand': kwargs.get('quantity_on_hand', UNKNOWN),
                            'sku': kwargs.get('sku_id', UNKNOWN),
                            'lead_time': kwargs.get('lead_time', UNKNOWN),
                            'unit_cost': kwargs.get('unit_cost', UNKNOWN),
                            'currency': kwargs.get('currency', UNKNOWN)
                            })
        return func(**temp_kwargs)

    return wrapper


def coroutine_decorator(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        coroutine = func(*args, **kwargs)
        coroutine.next()
        return coroutine

    return wrapper


def pickle_conifg(func):
    @wraps(func)
    def wrapper(**kwargs):
        with open(kwargs['file_path'], 'wb') as f:
            pickle.dump(kwargs['config'], f)



# Copyright (c) 2015-2016, The Authors and Contributors
# <see AUTHORS file>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
# following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
# following disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


from supplychainpy.reporting.extensions import db


def convert_datetime(value):
    """Deserialize datetime object into string"""
    if value is None:
        return None
    return [value.strftime("%d-%m-%y"), value.strftime("%H:%M:%S")]


class MasterSkuList(db.Model):
    __table_args__ = {'sqlite_autoincrement': True}
    id = db.Column(db.Integer(), primary_key=True)
    sku_id = db.Column(db.String(255))
    analysis = db.relationship("InventoryAnalysis", backref='sku', lazy='dynamic')
    simulation_summary_id = db.relationship("SimulationSummary", backref='sim', lazy='dynamic')


class Currency(db.Model):
    __table_args__ = {'sqlite_autoincrement': True}
    id = db.Column(db.Integer(), primary_key=True)
    fx = db.relationship("InventoryAnalysis", backref='currency', lazy='dynamic')
    currency_code = db.Column(db.String(3))
    country = db.Column(db.String(255))
    symbol = db.Column(db.String(255))


class TransactionLog(db.Model):
    __table_args__ = {'sqlite_autoincrement': True}
    id = db.Column(db.Integer(), primary_key=True)
    date = db.Column(db.DateTime())
    transaction = db.relationship("InventoryAnalysis", backref='log', lazy='dynamic')
    profile_recommendation = db.relationship("ProfileRecommendation", backref='profile', lazy='dynamic')


class InventoryAnalysis(db.Model):
    __table_args__ = {'sqlite_autoincrement': True}
    id = db.Column(db.Integer(), primary_key=True)
    sku_id = db.Column(db.Integer, db.ForeignKey('master_sku_list.id'))
    abc_xyz_classification = db.Column(db.String(2))
    standard_deviation = db.Column(db.Integer())
    safety_stock = db.Column(db.Integer())
    reorder_level = db.Column(db.Integer())
    economic_order_quantity = db.Column(db.Integer())
    demand_variability = db.Column(db.Integer())
    average_orders = db.Column(db.Float())
    shortages = db.Column(db.Integer())
    excess_stock = db.Column(db.Integer())
    reorder_quantity = db.Column(db.Integer())
    economic_order_variable_cost = db.Column(db.Float())
    unit_cost = db.Column(db.Numeric(12, 2))
    revenue = db.Column(db.Float())
    date = db.Column(db.DateTime())
    safety_stock_rank = db.Column(db.Integer())
    shortage_rank = db.Column(db.Integer())
    excess_cost = db.Column(db.Numeric(12, 2))
    percentage_contribution_revenue = db.Column(db.Float)
    excess_rank = db.Column(db.Integer())
    retail_price = db.Column(db.Numeric(18, 2))
    gross_profit_margin = db.Column(db.Numeric(12, 2))
    min_order = db.Column(db.Integer())
    safety_stock_cost = db.Column(db.Numeric(18, 2))
    revenue_rank = db.Column(db.Integer())
    markup_percentage = db.Column(db.Float())
    max_order = db.Column(db.Integer())
    shortage_cost = db.Column(db.Numeric(18, 2))
    quantity_on_hand = db.Column(db.Integer())
    currency_id = db.Column(db.Integer, db.ForeignKey('currency.id'))
    transaction_log_id = db.Column(db.Integer, db.ForeignKey('transaction_log.id'))
    orders_id = db.relationship("Orders", backref='demand', lazy='dynamic')
    forecast_id = db.relationship("Forecast", backref='forecasts', lazy='dynamic')
    forecast_breakdown_id = db.relationship("ForecastBreakdown", backref='estimates', lazy='dynamic')
    inventory_turns = db.Column(db.Float())
    traffic_light = db.Column(db.String(6))
    recommendation_id = db.relationship("Recommendations", backref='rec', lazy='dynamic')
    backlog = db.Column(db.Integer())

    @property
    def serialize(self):
        """Return object data in easily serializeable format"""
        return {
            'id': self.sku_id,
            'revenue': self.revenue,
            'classification': self.abc_xyz_classification,
            'eoq': self.economic_order_quantity,
            'Date': convert_datetime(self.date),
            'standard_deviation': self.standard_deviation,
            'safety_stock': self.safety_stock,
            'reorder_level': self.reorder_level,
            'economic_order_quantity': self.economic_order_quantity,
            'demand_variability': self.demand_variability,
            'average_orders': round(self.average_orders),
            'shortages': self.shortages,
            'excess_stock': self.excess_stock,
            'reorder_quantity': float(self.reorder_quantity),
            'economic_order_variable_cost': float(self.economic_order_variable_cost),
            'unit_cost': float(self.unit_cost),
            'safety_stock_rank': self.safety_stock_rank,
            'shortage_rank': self.shortage_rank,
            'excess_cost': float(self.excess_cost),
            'percentage_contribution_revenue': float(self.percentage_contribution_revenue),
            'excess_rank': self.excess_rank,
            'retail_price': float(self.retail_price),
            'gross_profit_margin': float(self.gross_profit_margin),
            'min_order': self.min_order,
            'safety_stock_cost': float(self.safety_stock_cost),
            'revenue_rank': self.revenue_rank,
            'markup_percentage': self.markup_percentage,
            'max_order': self.max_order,
            'shortage_cost': float(self.shortage_cost),
            'quantity_on_hand': self.quantity_on_hand,
            'inventory_turns': self.inventory_turns
        }


class Orders(db.Model):
    __table_args__ = {'sqlite_autoincrement': True}
    id = db.Column(db.Integer(), primary_key=True)
    analysis_id = db.Column(db.Integer, db.ForeignKey('inventory_analysis.id'))
    order_quantity = db.Column(db.Integer())
    rank = db.Column(db.Integer())


class ForecastType(db.Model):
    __table_args__ = {'sqlite_autoincrement': True}
    id = db.Column(db.Integer(), primary_key=True)
    type = db.Column(db.String(10))
    forecast_id = db.relationship("Forecast", backref='type', lazy='dynamic')
    forecast_statistic_id = db.relationship("ForecastStatistics", backref='details', lazy='dynamic')
    forecast_breakdown_id = db.relationship("ForecastBreakdown", backref='breakdown', lazy='dynamic')


class Forecast(db.Model):
    __table_args__ = {'sqlite_autoincrement': True}
    id = db.Column(db.Integer(), primary_key=True)
    analysis_id = db.Column(db.Integer, db.ForeignKey('inventory_analysis.id'))
    forecast_quantity = db.Column(db.Integer())
    forecast_type_id = db.Column(db.Integer, db.ForeignKey('forecast_type.id'))
    period = db.Column(db.Integer())
    create_date = db.Column(db.DateTime())


class ForecastStatistics(db.Model):
    __table_args__ = {'sqlite_autoincrement': True}
    id = db.Column(db.Integer(), primary_key=True)
    analysis_id = db.Column(db.Integer, db.ForeignKey('inventory_analysis.id'))
    forecast_type_id = db.Column(db.Integer, db.ForeignKey('forecast_type.id'))
    slope = db.Column(db.Float())
    p_value = db.Column(db.Float())
    test_statistic = db.Column(db.Float())
    slope_standard_error = db.Column(db.Float())
    intercept = db.Column(db.Float())
    standard_residuals = db.Column(db.Float())
    trending = db.Column(db.Boolean)
    mape = db.Column(db.Float())
    optimal_alpha = db.Column(db.Float())
    optimal_gamma = db.Column(db.Float())


class ForecastBreakdown(db.Model):
    __table_args__ = {'sqlite_autoincrement': True}
    id = db.Column(db.Integer(), primary_key=True)
    period = db.Column(db.Integer())
    analysis_id = db.Column(db.Integer, db.ForeignKey('inventory_analysis.id'))
    forecast_type_id = db.Column(db.Integer, db.ForeignKey('forecast_type.id'))
    level_estimates = db.Column(db.Float())
    trend = db.Column(db.Float())
    one_step_forecast = db.Column(db.Float())
    forecast_error = db.Column(db.Float())
    squared_error = db.Column(db.Float())
    regression = db.Column(db.Float())


class Recommendations(db.Model):
    __table_args__ = {'sqlite_autoincrement': True}
    id = db.Column(db.Integer(), primary_key=True)
    analysis_id = db.Column(db.Integer, db.ForeignKey('inventory_analysis.id'))
    statement = db.Column(db.Text())


class ProfileRecommendation(db.Model):
    __table_args__ = {'sqlite_autoincrement': True}
    id = db.Column(db.Integer(), primary_key=True)
    transaction_id = db.Column(db.Integer, db.ForeignKey('transaction_log.id'))
    statement = db.Column(db.Text())


class Simulation(db.Model):
    __table_args__ = {'sqlite_autoincrement': True}
    id = db.Column(db.Integer(), primary_key=True)
    date = db.Column(db.DateTime())
    runs = db.Column(db.Integer())
    sim_summary_id = db.relationship("SimulationSummary", backref='simulation', lazy='dynamic')


class SimulationSummary(db.Model):
    __table_args__ = {'sqlite_autoincrement': True}
    id = db.Column(db.Integer(), primary_key=True)
    sku_id = db.Column(db.Integer, db.ForeignKey('master_sku_list.id'))
    sim_id = db.Column(db.Integer, db.ForeignKey('simulation.id'))
    variance_opening_stock = db.Column(db.Integer())
    minimum_opening_stock = db.Column(db.Integer())
    maximum_closing_stock =  db.Column(db.Integer())
    maximum_quantity_sold = db.Column(db.Integer())
    maximum_backlog = db.Column(db.Integer())
    average_backlog = db.Column(db.Integer())
    minimum_closing_stock = db.Column(db.Integer())
    minimum_backlog = db.Column(db.Integer())
    minimum_quantity_sold = db.Column(db.Integer())
    average_quantity_sold = db.Column(db.Integer())
    maximum_opening_stock = db.Column(db.Integer())
    standard_deviation_quantity_sold = db.Column(db.Integer())
    standard_deviation_closing_stock = db.Column(db.Integer())
    average_closing_stock = db.Column(db.Integer())
    standard_deviation_backlog = db.Column(db.Integer())
    average_shortage_units = db.Column(db.Integer())



# Copyright (c) 2015-2016, The Authors and Contributors
# <see AUTHORS file>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
# following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
# following disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os

import flask
import logging
from flask import Blueprint
from flask import Flask
from flask import send_from_directory

from sqlalchemy import func, desc, asc

from supplychainpy.reporting.blueprints.models import (InventoryAnalysis,
                                                       Currency,
                                                       MasterSkuList, Recommendations, TransactionLog, Orders,
                                                       ForecastBreakdown, Forecast,
                                                       ForecastStatistics)
from supplychainpy.reporting.config.settings import ProdConfig
from supplychainpy.reporting.extensions import db, manager

log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())

app_dir = os.path.dirname(__file__, )
rel_path = '../uploads'
abs_file_path = os.path.abspath(os.path.join(app_dir, '..', rel_path))
UPLOAD_FOLDER = abs_file_path
app = Flask(__name__)
app.config.from_object(ProdConfig)
app.config['UPLOAD_FOLDER'] = UPLOAD_FOLDER
DASHBOARD_TEMPLATE_DIR = os.path.dirname(os.path.abspath(__file__))
dashboard_blueprint = Blueprint('dashboard', __name__, template_folder='templates')

manager.create_api(InventoryAnalysis, methods=['GET', 'POST', 'DELETE', 'PATCH'], allow_functions=True,
                   results_per_page=10, max_results_per_page=500)
manager.create_api(Currency, methods=['GET', 'POST', 'DELETE', 'PATCH'], allow_functions=True)
manager.create_api(Orders, methods=['GET', 'POST', 'DELETE', 'PATCH'], allow_functions=True)
manager.create_api(Forecast, methods=['GET', 'POST', 'DELETE', 'PATCH'], allow_functions=True)
manager.create_api(ForecastStatistics, methods=['GET', 'POST', 'DELETE', 'PATCH'], allow_functions=True)
manager.create_api(ForecastBreakdown, methods=['GET', 'POST', 'DELETE', 'PATCH'], allow_functions=True)
manager.create_api(MasterSkuList, methods=['GET', 'POST', 'DELETE', 'PATCH'], allow_functions=True)
manager.create_api(Recommendations, methods=['GET', 'POST', 'DELETE', 'PATCH'], allow_functions=True)



@dashboard_blueprint.route('/', methods=['GET'])
def dashboard():
    top_ten_shortages = db.session.query(InventoryAnalysis).order_by(desc(InventoryAnalysis.shortage_cost)).limit(10)
    largest_shortage = db.session.query(InventoryAnalysis).order_by(asc(InventoryAnalysis.shortage_rank)).first()
    total_shortages = db.session.query(InventoryAnalysis).order_by(desc(InventoryAnalysis.shortage_cost)).limit(10)
    total_shortages = total_shortages.with_entities(func.sum(InventoryAnalysis.shortage_cost)).scalar
    top_ten_excess = db.session.query(InventoryAnalysis).order_by(desc(InventoryAnalysis.excess_cost)).limit(10)
    largest_excess = db.session.query(InventoryAnalysis).order_by(asc(InventoryAnalysis.excess_rank)).first()
    total_excess = db.session.query(InventoryAnalysis).order_by(desc(InventoryAnalysis.excess_cost)).limit(10)
    total_excess = total_excess.with_entities(func.sum(InventoryAnalysis.excess_cost)).scalar
    currency = db.session.query(Currency).all()

    return flask.render_template('dashboard.html', shortages=top_ten_shortages, currency=currency,
                                 largest= largest_shortage, shortage= total_shortages, excess=top_ten_excess,
                                 largest_excess=largest_excess, total_excess=total_excess)


#@app.route('/upload/', methods=['POST', 'GET'])
#def upload_file():
#    """
#
#    Returns:
#
#    """
#    target = UPLOAD_FOLDER
#    form = DataForm()
#    if request.method == 'POST':
#        upload(request=request)
#        return "{}".format(request)
#
#    return flask.render_template('upload.html', form=form)


#@app.route('/settings', methods=['POST', 'GET'])
#def settings():
#    """Route to settings view.
#
#    Returns:
#
#    """
#    form = SettingsForm()
#    return flask.render_template('settings.html', form=form)


@dashboard_blueprint.route('/reporting/api/v1.0/sku_detail', methods=['GET'])
@dashboard_blueprint.route('/reporting/api/v1.0/sku_detail/<string:sku_id>', methods=['GET'])
def sku_detail(sku_id: str = None):
    """Route for restful sku detail, whole content limited by most recent date or individual sku

    Args:
        sku_id:

    Returns:

    """
    if sku_id is not None:
        inventory = db.session.query(InventoryAnalysis).filter(InventoryAnalysis.sku_id == sku_id).all()
    else:
        transaction_sub = db.session.query(db.func.max(TransactionLog.date))
        transaction_id = db.session.query(TransactionLog).filter(TransactionLog.date == transaction_sub).first()
        inventory = db.session.query(InventoryAnalysis).all()

    return flask.jsonify(json_list=[i.serialize for i in inventory])


@dashboard_blueprint.route('/sku_detail', methods=['GET'])
@dashboard_blueprint.route('/sku_detail/<string:sku_id>', methods=['GET'])
def sku(sku_id: str = None):
    """route for restful sku detail, whole content limited by most recent date or individual sku"""
    if sku_id is not None:
        current_transaction = db.session.query(TransactionLog).order_by(desc(TransactionLog.id)).first()
        sku = db.session.query(MasterSkuList).filter(MasterSkuList.id == sku_id).first()
        inventory = db.session.query(InventoryAnalysis).filter(InventoryAnalysis.transaction_log_id == current_transaction.id, InventoryAnalysis.sku_id == sku.id).all()
        inven = db.session.query(InventoryAnalysis.id).filter(InventoryAnalysis.sku_id == sku.id).first()
        orders = db.session.query(Orders).filter(Orders.analysis_id == inven.id).order_by(
            asc(Orders.rank)).all()
        forecast_breakdown = db.session.query(ForecastBreakdown).filter(ForecastBreakdown.analysis_id == inven.id)
        forecast = db.session.query(Forecast).filter(Forecast.analysis_id == inven.id).all()
        forecast_statistics = db.session.query(ForecastStatistics).filter(ForecastStatistics.analysis_id == inven.id)
        recommend = db.session.query(Recommendations.statement).filter(
            Recommendations.analysis_id == inven.id).all()
        cur = db.session.query(Currency).all()

    return flask.render_template('sku.html', inventory=inventory, orders=orders, breakdown=forecast_breakdown,
                                 forecast=forecast, statistics=forecast_statistics, recommendations=recommend,
                                 currency=cur)


@dashboard_blueprint.route('/reporting/api/v1.0/abc_summary', methods=['GET'])
@dashboard_blueprint.route('/reporting/api/v1.0/abc_summary/<string:classification>', methods=['GET'])
def get_classification_summary(classification: str = None):
    """route for restful summary of costs by abcxyz classifications"""
    if classification is not None:
        revenue_classification = db.session.query(InventoryAnalysis.abc_xyz_classification,
                                                  InventoryAnalysis.currency_id,
                                                  Currency.currency_code,
                                                  func.sum(InventoryAnalysis.revenue).label('total_revenue'),
                                                  func.sum(InventoryAnalysis.shortage_cost).label('total_shortages'),
                                                  func.sum(InventoryAnalysis.excess_cost).label('total_excess')
                                                  ).join(Currency).filter(
            InventoryAnalysis.abc_xyz_classification == classification).group_by(
            InventoryAnalysis.abc_xyz_classification).all()
    else:
        revenue_classification = db.session.query(InventoryAnalysis.abc_xyz_classification,
                                                  InventoryAnalysis.currency_id,
                                                  Currency.currency_code,
                                                  func.sum(InventoryAnalysis.revenue).label('total_revenue'),
                                                  func.sum(InventoryAnalysis.shortage_cost).label('total_shortages'),
                                                  func.sum(InventoryAnalysis.excess_cost).label('total_excess')
                                                  ).join(Currency).group_by(
            InventoryAnalysis.abc_xyz_classification).all()

    return flask.jsonify(json_list=[i for i in revenue_classification])


@dashboard_blueprint.route('/abcxyz/<string:classification>')
def abxyz(classification: str = None):
    abc = db.session.query(InventoryAnalysis
                           ).filter(
        InventoryAnalysis.abc_xyz_classification == classification).all()

    msk = db.session.query(MasterSkuList).all()
    cur = db.session.query(Currency).all()

    return flask.render_template('abcxyz.html', inventory=abc, mks=msk, currency=cur)


@dashboard_blueprint.route('/reporting/api/v1.0/top_shortages', methods=['GET'])
@dashboard_blueprint.route('/reporting/api/v1.0/top_shortages/<int:rank>', methods=['GET'])
@dashboard_blueprint.route('/reporting/api/v1.0/top_shortages/<string:classification>', methods=['GET'])
def top_shortages(rank: int = 10, classification: str = None):
    """ route for restful summary of top skus """
    if classification is not None:
        revenue_classification = db.session.query(InventoryAnalysis.abc_xyz_classification,
                                                  InventoryAnalysis.sku_id,
                                                  InventoryAnalysis.shortage_cost,
                                                  InventoryAnalysis.quantity_on_hand,
                                                  InventoryAnalysis.percentage_contribution_revenue,
                                                  InventoryAnalysis.revenue_rank,
                                                  InventoryAnalysis.shortages,
                                                  InventoryAnalysis.shortage_rank,
                                                  InventoryAnalysis.average_orders,
                                                  InventoryAnalysis.safety_stock,
                                                  InventoryAnalysis.reorder_level
                                                  ).filter(
            InventoryAnalysis.abc_xyz_classification == classification).order_by(
            desc(InventoryAnalysis.shortage_cost)).limit(rank)
    else:
        revenue_classification = db.session.query(InventoryAnalysis.abc_xyz_classification,
                                                  InventoryAnalysis.sku_id,
                                                  InventoryAnalysis.shortage_cost,
                                                  InventoryAnalysis.quantity_on_hand,
                                                  InventoryAnalysis.percentage_contribution_revenue,
                                                  InventoryAnalysis.revenue_rank,
                                                  InventoryAnalysis.shortages,
                                                  InventoryAnalysis.shortage_rank,
                                                  InventoryAnalysis.average_orders,
                                                  InventoryAnalysis.safety_stock,
                                                  InventoryAnalysis.reorder_level
                                                  ).order_by(desc(InventoryAnalysis.shortage_cost)).limit(rank)

    return flask.jsonify(json_list=[i for i in revenue_classification])


@dashboard_blueprint.route('/reporting/api/v1.0/top_excess', methods=['GET'])
@dashboard_blueprint.route('/reporting/api/v1.0/top_excess/<int:rank>', methods=['GET'])
@dashboard_blueprint.route('/reporting/api/v1.0/top_excess/<string:classification>', methods=['GET'])
def top_excess(rank: int = 10, classification: str = None):
    """ route for restful summary of top skus """
    if classification is not None:
        revenue_classification = db.session.query(InventoryAnalysis.abc_xyz_classification,
                                                  InventoryAnalysis.sku_id,
                                                  InventoryAnalysis.excess_cost,
                                                  InventoryAnalysis.quantity_on_hand,
                                                  InventoryAnalysis.percentage_contribution_revenue,
                                                  InventoryAnalysis.revenue_rank,
                                                  InventoryAnalysis.excess_stock,
                                                  InventoryAnalysis.average_orders,
                                                  InventoryAnalysis.safety_stock,
                                                  InventoryAnalysis.reorder_level
                                                  ).filter(
            InventoryAnalysis.abc_xyz_classification == classification).order_by(
            desc(InventoryAnalysis.excess_cost)).limit(rank)
    else:
        revenue_classification = db.session.query(InventoryAnalysis.abc_xyz_classification,
                                                  InventoryAnalysis.sku_id,
                                                  InventoryAnalysis.excess_cost,
                                                  InventoryAnalysis.quantity_on_hand,
                                                  InventoryAnalysis.percentage_contribution_revenue,
                                                  InventoryAnalysis.revenue_rank,
                                                  InventoryAnalysis.excess_stock,
                                                  InventoryAnalysis.average_orders,
                                                  InventoryAnalysis.safety_stock,
                                                  InventoryAnalysis.reorder_level
                                                  ).order_by(desc(InventoryAnalysis.excess_cost)).limit(rank)

    return flask.jsonify(json_list=[i for i in revenue_classification])


@dashboard_blueprint.route('/api/total_inventory', methods=['GET'])
def total_inventory():
    revenue_classification = db.session.query(
        InventoryAnalysis.sku_id,
        InventoryAnalysis.unit_cost,
        InventoryAnalysis.quantity_on_hand,
        InventoryAnalysis.transaction_log_id).all()

    return flask.jsonify(json_list=[i for i in revenue_classification])


@dashboard_blueprint.route('/reporting/api/v1.0/currency', methods=['GET'])
def currency():
    pass
    #  currency_code = db.session.query(Currency.id).filter(Currency.currency_code == item['currency']).first()


# @app.route('/upload/', methods=('GET', 'POST'))
# def upload():
#    form = DataForm()
#    upload_data = DataUpload()
#
    #    if request.method == 'POST':
#        file = request.files['data']
#        if file and allowed_file(file.filename):
#            filename = secure_filename(file.filename)
#            file.save(os.path.join(app.config['UPLOAD_FOLDER'], filename))
#            # upload_data.data = filename
#            # upload_data.date = datetime.now()
#            # db.session.add(upload_data)
#            # db.session.commit()
#
#            return redirect(url_for('uploaded_file',
#                                    filename=filename))
#
#
#    return flask.render_template('upload.html', form=form)



@dashboard_blueprint.route('/uploads/<filename>')
def uploaded_file(filename):
    return send_from_directory(app.config['UPLOAD_FOLDER'],
                               filename)




if __name__ == '__main__':
    app.run()



import flask
from flask import Blueprint
from flask import Flask

from supplychainpy.reporting.blueprints.models import InventoryAnalysis, Currency
from supplychainpy.reporting.config.settings import DevConfig
from supplychainpy.reporting.extensions import db

rawdata_blueprint = Blueprint('rawdata', __name__, template_folder='templates')


app = Flask(__name__)
app.config.from_object(DevConfig)

@rawdata_blueprint.route('/data')
def rawdata():
    """Route for the raw data from the analysis.

    Returns:

    """
    inventory = db.session.query(InventoryAnalysis).all()
    cur = db.session.query(Currency).all()

    return flask.render_template('rawdata.html', inventory=inventory, currency=cur)




# Copyright (c) 2015-2016, The Authors and Contributors
# <see AUTHORS file>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
# following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
# following disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import flask
from flask import Blueprint, render_template
from supplychainpy.bot.dash import ChatBot

bot_blueprint = Blueprint('bot', __name__, template_folder='templates')


@bot_blueprint.route('/bot')
def bot():
    """the route for Dash the chat bot."""
    return render_template('bot.html')


@bot_blueprint.route('/bot/lexicon')
def lexicon():
    """The route for displaying examples for interacting with the chat bot.

    Returns:
        html

    """
    return render_template('lexicon.html')


@bot_blueprint.route('/chat/<string:message>', methods=['GET'])
def chat(message: str = None):
    """ Rest api for chat bot

    Args:
        message: User query to chat bot.

    Returns:
        json:   Response from chat bot.

    """
    dash = ChatBot()
    response = dash.chat_machine(message=message)

    return flask.jsonify(json_list=response)
